"""Test damerau levenshtein dist norm that is exposed in the context class and helper functions."""
import os
from datetime import datetime
import pytest
import static_frame as sf
import aqudem
from aqudem.damerau_levenshtein_helper import (_map_strings_to_letters,
                                               _damerau_opt_levenshtein_dist,
                                               _levenshtein_distancy_by_case,
                                               _damerau_levenshtein_distancy_by_case)


def test_basic_map_strings() -> None:
    res1, res2 = _map_strings_to_letters(["i", "think", "i", "am", "think", "think", "am", "i"],
                                         ["therefore", "therefore", "am",
                                          "i", "i", "am", "therefore", "am"])
    assert res1 == "ABACBBCA"
    assert res2 == "DDCAACDC"


def test_different_lengths_map_strings() -> None:
    res1, res2 = _map_strings_to_letters(["i", "think", "i", "am"],
                                         ["therefore", "therefore", "am", "i",
                                          "i", "am", "therefore", "am"])
    assert res1 == "ABAC"
    assert res2 == "DDCAACDC"


def test_empty_lengths_map_strings() -> None:
    res1, res2 = _map_strings_to_letters([],
                                         ["therefore", "therefore", "am", "i",
                                          "i", "am", "therefore", "am"])
    assert res1 == ""
    assert res2 == "AABCCBAB"
    res1, res2 = _map_strings_to_letters(["i", "think", "i", "am"],
                                         [])
    assert res1 == "ABAC"
    assert res2 == ""


def test_basic_damerau_opt_levenshtein_dist_norm() -> None:
    gt = sf.FrameHE.from_dict({
        "case:concept:name": ["ExampleTrace", "ExampleTrace", "ExampleTrace", "ExampleTrace"],
        "lifecycle:transition": ["start", "complete", "start", "complete"],
        "concept:name": ["A", "A", "B", "B"],
        "time:timestamp": [datetime(2021, 1, 1, 0, 0, 0),
                           datetime(2021, 1, 1, 0, 10, 0),
                           datetime(2021, 1, 1, 0, 15, 0),
                           datetime(2021, 1, 1, 0, 20, 0)]
    })
    det = sf.FrameHE.from_dict({
        "case:concept:name": ["ExampleTrace", "ExampleTrace", "ExampleTrace", "ExampleTrace"],
        "lifecycle:transition": ["start", "complete", "start", "complete"],
        "concept:name": ["A", "A", "B", "B"],
        "time:timestamp": [datetime(2021, 1, 1, 0, 0, 0),
                           datetime(2021, 1, 1, 0, 10, 0),
                           datetime(2021, 1, 1, 0, 15, 0),
                           datetime(2021, 1, 1, 0, 20, 0)]
    })
    res = _damerau_opt_levenshtein_dist(gt, det)
    assert res[1] == 0.0


def test_swap_damerau_opt_levenshtein_dist_norm() -> None:
    gt = sf.FrameHE.from_dict({
        "case:concept:name": ["ExampleTrace", "ExampleTrace"],
        "lifecycle:transition": ["start", "start"],
        "concept:name": ["A", "B"],
        "time:timestamp": [datetime(2021, 1, 1, 0, 0, 0),
                           datetime(2021, 1, 1, 0, 10, 0)]
    })
    det = sf.FrameHE.from_dict({
        "case:concept:name": ["ExampleTrace", "ExampleTrace"],
        "lifecycle:transition": ["start", "start"],
        "concept:name": ["B", "A"],
        "time:timestamp": [datetime(2021, 1, 1, 0, 0, 0),
                           datetime(2021, 1, 1, 0, 10, 0)]
    })
    res = _damerau_opt_levenshtein_dist(gt, det)
    assert res[1] == 0.5


def test_change_damerau_opt_levenshtein_dist_norm() -> None:
    gt = sf.FrameHE.from_dict({
        "case:concept:name": ["ExampleTrace", "ExampleTrace"],
        "lifecycle:transition": ["start", "start"],
        "concept:name": ["A", "B"],
        "time:timestamp": [datetime(2021, 1, 1, 0, 0, 0),
                           datetime(2021, 1, 1, 0, 10, 0)]
    })
    det = sf.FrameHE.from_dict({
        "case:concept:name": ["ExampleTrace", "ExampleTrace"],
        "lifecycle:transition": ["start", "start"],
        "concept:name": ["B", "A"],
        "time:timestamp": [datetime(2021, 1, 1, 0, 0, 0),
                           datetime(2021, 1, 1, 0, 10, 0)]
    })
    res = _damerau_opt_levenshtein_dist(gt, det)
    assert res[1] == 0.5
    res = _damerau_opt_levenshtein_dist(gt, det, metr_type="lev")
    assert res[1] == 1.0


def test_change_damerau_opt_levenshtein_dist() -> None:
    gt = sf.FrameHE.from_dict({
        "case:concept:name": ["ExampleTrace", "ExampleTrace"],
        "lifecycle:transition": ["start", "start"],
        "concept:name": ["A", "B"],
        "time:timestamp": [datetime(2021, 1, 1, 0, 0, 0),
                           datetime(2021, 1, 1, 0, 10, 0)]
    })
    det = sf.FrameHE.from_dict({
        "case:concept:name": ["ExampleTrace", "ExampleTrace"],
        "lifecycle:transition": ["start", "start"],
        "concept:name": ["B", "A"],
        "time:timestamp": [datetime(2021, 1, 1, 0, 0, 0),
                           datetime(2021, 1, 1, 0, 10, 0)]
    })
    res = _damerau_opt_levenshtein_dist(gt, det)
    assert res[0] == 1
    res = _damerau_opt_levenshtein_dist(gt, det, metr_type="lev")
    assert res[0] == 2


def test_change_damerau_opt_levenshtein() -> None:
    gt = sf.FrameHE.from_dict({
        "case:concept:name": ["ExampleTrace", "ExampleTrace", "ExampleTrace2",
                              "ExampleTrace2", "ExampleTrace2", "ExampleTrace2"],
        "lifecycle:transition": ["start", "start", "start", "start", "start", "start"],
        "concept:name": ["A", "B", "A", "B", "A", "B"],
        "time:timestamp": [datetime(2021, 1, 1, 0, 0, 0),
                           datetime(2021, 1, 1, 0, 10, 0),
                           datetime(2021, 1, 1, 0, 0, 0),
                           datetime(2021, 1, 1, 0, 10, 0),
                           datetime(2021, 1, 1, 0, 20, 0),
                           datetime(2021, 1, 1, 0, 30, 0)]
    })
    det = sf.FrameHE.from_dict({
        "case:concept:name": ["ExampleTrace", "ExampleTrace", "ExampleTrace2",
                              "ExampleTrace2", "ExampleTrace2", "ExampleTrace2"],
        "lifecycle:transition": ["start", "start", "start", "start", "start", "start"],
        "concept:name": ["B", "A", "B", "A", "C", "C"],
        "time:timestamp": [datetime(2021, 1, 1, 0, 0, 0),
                           datetime(2021, 1, 1, 0, 10, 0),
                           datetime(2021, 1, 1, 0, 0, 0),
                           datetime(2021, 1, 1, 0, 10, 0),
                           datetime(2021, 1, 1, 0, 20, 0),
                           datetime(2021, 1, 1, 0, 30, 0)]
    })
    res = _damerau_levenshtein_distancy_by_case(gt, det, case_id="ExampleTrace")
    assert res == (1, 0.5)
    res = _damerau_levenshtein_distancy_by_case(gt, det, case_id="ExampleTrace2")
    assert res == (3, 0.75)
    res = _damerau_levenshtein_distancy_by_case(gt, det, case_id="*")
    assert res == (2, 0.625)
    res = _levenshtein_distancy_by_case(gt, det, case_id="ExampleTrace")
    assert res == (2, 1.0)
    res = _levenshtein_distancy_by_case(gt, det, case_id="ExampleTrace2")
    assert res == (3, 0.75)
    res = _levenshtein_distancy_by_case(gt, det, case_id="*")
    assert res == (2.5, 0.875)


def test_change_all_damerau_opt_levenshtein_dist_norm() -> None:
    gt = sf.FrameHE.from_dict({
        "case:concept:name": ["ExampleTrace", "ExampleTrace"],
        "lifecycle:transition": ["start", "start"],
        "concept:name": ["A", "B"],
        "time:timestamp": [datetime(2021, 1, 1, 0, 0, 0),
                           datetime(2021, 1, 1, 0, 10, 0)]
    })
    det = sf.FrameHE.from_dict({
        "case:concept:name": ["ExampleTrace", "ExampleTrace"],
        "lifecycle:transition": ["start", "start"],
        "concept:name": ["C", "C"],
        "time:timestamp": [datetime(2021, 1, 1, 0, 0, 0),
                           datetime(2021, 1, 1, 0, 10, 0)]
    })
    res = _damerau_opt_levenshtein_dist(gt, det)
    assert res[1] == 1.0
    gt = sf.FrameHE.from_dict({
        "case:concept:name": ["ExampleTrace", "ExampleTrace"],
        "lifecycle:transition": ["start", "start"],
        "concept:name": ["A", "B"],
        "time:timestamp": [datetime(2021, 1, 1, 0, 0, 0),
                           datetime(2021, 1, 1, 0, 10, 0)]
    })
    det = sf.FrameHE.from_dict({
        "case:concept:name": [],
        "lifecycle:transition": [],
        "concept:name": [],
        "time:timestamp": []
    })
    res = _damerau_opt_levenshtein_dist(gt, det)
    assert res[1] == 1.0


def test_empty_damerau_opt_levenshtein_dist_norm() -> None:
    gt = sf.FrameHE.from_dict({
        "case:concept:name": ["ExampleTrace", "ExampleTrace"],
        "lifecycle:transition": ["start", "start"],
        "concept:name": ["A", "B"],
        "time:timestamp": [datetime(2021, 1, 1, 0, 0, 0),
                           datetime(2021, 1, 1, 0, 10, 0)]
    })
    det = sf.FrameHE.from_dict({
        "case:concept:name": [],
        "lifecycle:transition": [],
        "concept:name": [],
        "time:timestamp": []
    })
    res = _damerau_opt_levenshtein_dist(gt, det)
    assert res[1] == 1.0
    gt = sf.FrameHE.from_dict({
        "case:concept:name": [],
        "lifecycle:transition": [],
        "concept:name": [],
        "time:timestamp": []
    })
    det = sf.FrameHE.from_dict({
        "case:concept:name": [],
        "lifecycle:transition": [],
        "concept:name": [],
        "time:timestamp": []
    })
    res = _damerau_opt_levenshtein_dist(gt, det)
    assert res[1] == 0.0


def test_context_damerau_levenshtein() -> None:
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.damerau_levenshtein_distance(case_id="ExampleTrace1")
    assert res == (2, 0.3333)
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.damerau_levenshtein_distance(case_id="ExampleTrace2")
    assert res == (1, 0.3333)
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.damerau_levenshtein_distance()
    assert res == (1.5, 0.3333)


def test_context_levenshtein() -> None:
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.levenshtein_distance(case_id="ExampleTrace2")
    assert res == (2, 0.6667)
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.levenshtein_distance(case_id="ExampleTrace1")
    assert res == (2, 0.3333)
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.levenshtein_distance()
    assert res == (2.0, 0.5)


def test_wrong_metr_type() -> None:
    with pytest.raises(ValueError):
        _damerau_opt_levenshtein_dist(sf.FrameHE(), sf.FrameHE(), metr_type="wrong")



""" Helper functions to calculate the Event Analysis (EA) metrics. """
#pylint: disable=too-many-statements, too-many-locals, too-many-branches
from dataclasses import dataclass, fields
from collections import Counter
from functools import cached_property, lru_cache
from typing import Union, Any, Dict
import numpy as np
import static_frame as sf

from .utils import (_case_activity_level_metric_pre_check, BUG_REPORT_CTA,
                    _generate_activity_metric_list, _get_case_in_filtered_logs)
from .ward_helper import _generate_segment_scores


# pylint: disable=too-many-instance-attributes
@dataclass(frozen=True)
class EventAnalysis:
    """Data class to hold the EA metrics.

    Regarding the ground truth events: d, f, fm, m.
    Regarding both the ground truth and detected events: c.
    Regarding the (d)etected events: md, fmd, fd, id.
    If result of aggregated request, the values represent the average number of events
    over the relevant case-activity pairs.
    Relative metrics are available as properties.

    :param d: Deletions
    :param f: Fragmentations
    :param fm: Fragmentation and merge
    :param m: Merges
    :param c: Correct
    :param md: Merges
    :param fmd: Fragmentation and merge
    :param fd: Fragmentations
    :param id: Insertions
    """
    d: Union[int, float]
    f: Union[int, float]
    fm: Union[int, float]
    m: Union[int, float]
    c: Union[int, float]
    md: Union[int, float]
    fmd: Union[int, float]
    fd: Union[int, float]
    id: Union[int, float]

    @cached_property
    def total_gt_events(self) -> Union[int, float]:
        """Get the total number of ground truth events."""
        return round(self.d + self.f + self.fm + self.m + (self.c / 2), 4)

    @cached_property
    def total_det_events(self) -> Union[int, float]:
        """Get the total number of detected events."""
        return round((self.c / 2) + self.md + self.fmd + self.fd + self.id, 4)

    @cached_property
    def correct_events_per_log(self) -> Union[int, float]:
        """Get the total number of correct events per log."""
        return round(self.c / 2, 4)

    @cached_property
    def dr(self) -> float:
        """Get the deletion rate.
        Ratio of deletions to total ground truth events."""
        return (round(self.d / self.total_gt_events, 4)
                if self.total_gt_events > 0
                else 0)

    @cached_property
    def fr(self) -> float:
        """Get the fragmentation rate.
        Ratio of fragmentations to total ground truth events."""
        return (round(self.f / self.total_gt_events, 4)
                if self.total_gt_events > 0
                else 0)

    @cached_property
    def fmr(self) -> float:
        """Get the fragmentation and merge rate.
        Ratio of fragmentation and merge to total ground truth events."""
        return (round(self.fm / self.total_gt_events, 4)
                if self.total_gt_events > 0
                else 0)

    @cached_property
    def mr(self) -> float:
        """Get the merge rate.
        Ratio of merges to total ground truth events."""
        return (round(self.m / self.total_gt_events, 4)
                if self.total_gt_events > 0
                else 0)

    @cached_property
    def cr_gt(self) -> float:
        """Get the correct rate.
        Ratio of correct events per log to total ground truth events."""
        return (round(self.correct_events_per_log / self.total_gt_events, 4)
                if self.total_gt_events > 0
                else 0)

    @cached_property
    def mdr(self) -> float:
        """Get the merging rate.
        Ratio of merging to total detected events."""
        return (round(self.md / self.total_det_events, 4)
                if self.total_det_events > 0
                else 0)

    @cached_property
    def fmdr(self) -> float:
        """Get the fragmentating and merging rate.
        Ratio of fragmentating and merging to total detected events."""
        return (round(self.fmd / self.total_det_events, 4)
                if self.total_det_events > 0
                else 0)

    @cached_property
    def fdr(self) -> float:
        """Get the fragmentating rate.
        Ratio of fragmentating to total detected events."""
        return (round(self.fd / self.total_det_events, 4)
                if self.total_det_events > 0
                else 0)

    @cached_property
    def idr(self) -> float:
        """Get the insertion rate.
        Ratio of insertion to total detected events."""
        return (round(self.id / self.total_det_events, 4)
                if self.total_det_events > 0
                else 0)

    @cached_property
    def cr_det(self) -> float:
        """Get the correct rate.
        Ratio of correct events per log to total detected events."""
        return (round(self.correct_events_per_log / self.total_det_events, 4)
                if self.total_det_events > 0
                else 0)

    @cached_property
    def true_positives(self) -> Union[int, float]:
        """Get the number of true/correct event detections."""
        return round(self.correct_events_per_log, 4)

    @cached_property
    def false_positives(self) -> Union[int, float]:
        """Get the number of false (not correct) event detections."""
        result  = round(self.md + self.fmd + self.fd + self.id, 4)
        if not abs(result - (self.total_det_events - self.correct_events_per_log)) < 0.01:
            raise ValueError(f"False positives calculation error. "
                             f"{BUG_REPORT_CTA}")
        return result

    @cached_property
    def false_negatives(self) -> Union[int, float]:
        """Get the number of ground truth events that have not been categorized as correct."""
        result = round(self.d + self.f + self.fm + self.m, 4)
        if not abs(result - (self.total_gt_events - self.correct_events_per_log)) < 0.01:
            raise ValueError(f"False negatives calculation error. "
                             f"{BUG_REPORT_CTA}")
        return result


    @cached_property
    def precision(self) -> float:
        """Get the precision.
        Ratio of true positives to all positive predictions."""
        return (round(self.true_positives / (self.true_positives + self.false_positives), 4)
                if self.true_positives + self.false_positives > 0
                else 0)

    @cached_property
    def recall(self) -> float:
        """Get the recall.
        Ratio of true positives to the number of all positives in ground truth."""
        return (round(self.true_positives / (self.true_positives + self.false_negatives), 4)
                if self.true_positives + self.false_negatives > 0
                else 0)

    @cached_property
    def f1(self) -> float:
        """Get the F1 score.
        Harmonic mean of precision and recall."""
        return (round((2 * self.precision * self.recall) / (self.precision + self.recall), 4)
                if self.precision + self.recall > 0
                else 0)


@lru_cache
def _event_analysis(gt: sf.FrameHE,
                    det: sf.FrameHE,
                    start_end_per_case: sf.SeriesHE[sf.Index[np.str_], Any],
                    _: float = -1.0) -> EventAnalysis:
    """Calculate the absolute EA metrics.

    Assume that the logs are filtered by activity and case id.
    :param sf.FrameHE gt: The ground truth event log.
    :param sf.FrameHE det: The detected event log.
    :param start_end_by_case: The start and end times of the cases.
    :return: The absolute EA metrics.
    """
    case_id = _get_case_in_filtered_logs(gt, det)
    _case_activity_level_metric_pre_check(gt, det)
    segment_scores = _generate_segment_scores(gt,
                                              det,
                                              start_end_per_case[case_id][0],
                                              start_end_per_case[case_id][1])
    proc_gt_rows = []
    proc_det_rows = []
    final_gt_rows = []
    final_det_rows = []
    for i in range(0, len(gt), 2):
        if (gt.iloc[i]["lifecycle:transition"] != "start"
            or gt.iloc[i + 1]["lifecycle:transition"] != "complete"):
            raise ValueError(f"Invalid log row ordering. {BUG_REPORT_CTA}")
        event_start = gt.iloc[i]["time:timestamp"]
        event_end = gt.iloc[i + 1]["time:timestamp"]
        proc_gt_rows.append({
            "start": event_start,
            "end": event_end,
            "types": []
        })
    for i in range(0, len(det), 2):
        if (det.iloc[i]["lifecycle:transition"] != "start"
            or det.iloc[i + 1]["lifecycle:transition"] != "complete"):
            raise ValueError(f"Invalid log row ordering. {BUG_REPORT_CTA}")
        event_start = det.iloc[i]["time:timestamp"]
        event_end = det.iloc[i + 1]["time:timestamp"]
        proc_det_rows.append({
            "start": event_start,
            "end": event_end,
            "types": []
        })
    for gt_event in proc_gt_rows:
        contained_segment_scores = segment_scores.loc[
            (segment_scores["start"] >= gt_event["start"])
            & (segment_scores["end"] <= gt_event["end"])]
        equal_segment_scores = segment_scores.loc[
            (segment_scores["start"] == gt_event["start"])
            & (segment_scores["end"] == gt_event["end"])]
        if (len(equal_segment_scores) == 1
            and equal_segment_scores.iloc[0]["type"] == "D"):
            gt_event["types"].append("D")
        elif len(contained_segment_scores.loc[contained_segment_scores["type"] == "F"]) > 0:
            gt_event["types"].append("F")
    for det_event in proc_det_rows:
        contained_segment_scores = segment_scores.loc[
            (segment_scores["start"] >= det_event["start"])
            & (segment_scores["end"] <= det_event["end"])]
        equal_segment_scores = segment_scores.loc[
            (segment_scores["start"] == det_event["start"])
            & (segment_scores["end"] == det_event["end"])]
        if (len(equal_segment_scores) == 1
            and equal_segment_scores.iloc[0]["type"] == "I"):
            det_event["types"].append("I'")
        elif len(contained_segment_scores.loc[contained_segment_scores["type"] == "M"]) > 0:
            det_event["types"].append("M'")
    for gt_event in proc_gt_rows:
        overlapping_det_events = [det_event for det_event in proc_det_rows
                          if (gt_event["end"] >= det_event["start"] >= gt_event["start"])
                          or (gt_event["start"] <= det_event["end"] <= gt_event["end"])
                          or ((det_event["start"] <= gt_event["start"] <= det_event["end"])
                              and (det_event["start"] <= gt_event["end"] <= det_event["end"]))]
        if any("M'" in det_event["types"] for det_event in overlapping_det_events):
            gt_event["types"].append("M")
    for det_event in proc_det_rows:
        overlapping_gt_events = [gt_event for gt_event in proc_gt_rows
                         if (det_event["end"] >= gt_event["start"] >= det_event["start"])
                         or (det_event["start"] <= gt_event["end"] <= det_event["end"])
                         or ((gt_event["start"] <= det_event["start"] <= gt_event["end"])
                             and (gt_event["start"] <= det_event["end"] <= gt_event["end"]))]
        if any("F" in gt_event["types"] for gt_event in overlapping_gt_events):
            det_event["types"].append("F'")
    for gt_event in proc_gt_rows:
        if len(gt_event["types"]) == 0:
            final_gt_rows.append({
                "start": gt_event["start"],
                "end": gt_event["end"],
                "type": "C"
            })
        elif "D" in gt_event["types"]:
            final_gt_rows.append({
                "start": gt_event["start"],
                "end": gt_event["end"],
                "type": "D"
            })
        elif "F" in gt_event["types"] and "M" in gt_event["types"]:
            final_gt_rows.append({
                "start": gt_event["start"],
                "end": gt_event["end"],
                "type": "FM"
            })
        elif "F" in gt_event["types"]:
            final_gt_rows.append({
                "start": gt_event["start"],
                "end": gt_event["end"],
                "type": "F"
            })
        elif "M" in gt_event["types"]:
            final_gt_rows.append({
                "start": gt_event["start"],
                "end": gt_event["end"],
                "type": "M"
            })
    for det_event in proc_det_rows:
        if len(det_event["types"]) == 0:
            final_det_rows.append({
                "start": det_event["start"],
                "end": det_event["end"],
                "type": "C"
            })
        elif "I'" in det_event["types"]:
            final_det_rows.append({
                "start": det_event["start"],
                "end": det_event["end"],
                "type": "I'"
            })
        elif "F'" in det_event["types"] and "M'" in det_event["types"]:
            final_det_rows.append({
                "start": det_event["start"],
                "end": det_event["end"],
                "type": "FM'"
            })
        elif "F'" in det_event["types"]:
            final_det_rows.append({
                "start": det_event["start"],
                "end": det_event["end"],
                "type": "F'"
            })
        elif "M'" in det_event["types"]:
            final_det_rows.append({
                "start": det_event["start"],
                "end": det_event["end"],
                "type": "M'"
            })
    gt_types = Counter(row["type"] for row in final_gt_rows)
    det_types = Counter(row["type"] for row in final_det_rows)
    if gt_types["C"] != det_types["C"]:
        raise ValueError(f"The number of correct events in the ground truth and detected logs "
                         f"must be equal. {BUG_REPORT_CTA}")
    return EventAnalysis(
        d=gt_types["D"],
        f=gt_types["F"],
        fm=gt_types["FM"],
        m=gt_types["M"],
        c=gt_types["C"] + det_types["C"],
        md=det_types["M'"],
        fmd=det_types["FM'"],
        fd=det_types["F'"],
        id=det_types["I'"]
    )


def _event_analysis_by_activity_case(gt: sf.FrameHE,
                                     det: sf.FrameHE,
                                     activity: str,
                                     case_id: str,
                                     start_end_per_case: \
                                        sf.SeriesHE[sf.Index[np.str_], Any]) \
                                            -> EventAnalysis:
    """Calculate the EA metrics for a given activity and case id.

    If both the activity and case id is given as "*", the EA metrics will 
    be calculated and summed up for all activity-case pairs.

    :param sf.FrameHE gt: The ground truth event log.
    :param sf.FrameHE det: The detected event log.
    :param str activity: The activity name.
        If "*" is passed, the EA metrics will be calculated and summed up for all activities.
    :param str case_id: The case id.
        If "*" is passed, the EA metrics will be calculated and summed up for all cases.
    :param start_end_per_case: The start and end times of the cases.
    :return: The EA metrics.
    """
    all_metrics = _generate_activity_metric_list(gt=gt,
                                                 det=det,
                                                 case_id=case_id,
                                                 activity_name=activity,
                                                 start_end_per_case=start_end_per_case,
                                                 metric=_event_analysis)
    sum_ea_dict: Dict[str, Union[float, int]] = {
        "d": 0,
        "f": 0,
        "fm": 0,
        "m": 0,
        "c": 0,
        "md": 0,
        "fmd": 0,
        "fd": 0,
        "id": 0
    }
    for ea_metr in all_metrics:
        for field in fields(EventAnalysis):
            sum_ea_dict[field.name] += getattr(ea_metr, field.name)

    return EventAnalysis(**sum_ea_dict)



"""This module contains general utility functionality for the aqudem library."""
from typing import List, Any, Callable, Dict

import numpy as np
import static_frame as sf

BUG_REPORT_CTA = "This is most likely a bug. Please report the issue."


class Error(Exception):  # will be used as aqudem.Error
    """Base class for all of this library's exceptions."""


class XESMissingTraceNameAttribute(Error):
    """The trace needs to be named."""


class XESSamplingFreqError(Error):
    """The trace is missing the sampling frequency or there is an inconsistency."""


class XESMissingTimestamp(Error):
    """An event is missing the timestamp attribute."""


class XESMissingActivityName(Error):
    """An event is missing the activity name attribute."""


class XESMissingActivityInstance(Error):
    """An event is missing the activity instance attribute."""


class XESIncorrectLifecycleTransitionError(Error):
    """The lifecycle transition is not correct."""


def _validate_xes_dataframe_before_processing(xes_ground_truth: sf.FrameHE,
                                              xes_detected: sf.FrameHE) -> None:
    """Check if the XES logs are in a valid format.

    :param xes_ground_truth: The ground truth XES log.
    :param xes_detected: The detected XES log.
    :returns: None"""
    if ("case:sampling_freq" not in xes_detected.columns
        or xes_detected["case:sampling_freq"].count(unique=True) != 1):
        raise XESSamplingFreqError("All XES traces in the log must have "
                                   "a 'sampling_freq' attribute.")
    for xes_frame in [xes_ground_truth, xes_detected]:
        if ("case:concept:name" not in xes_frame.columns
            or xes_frame["case:concept:name"].isna().any()):
            raise XESMissingTraceNameAttribute("The trace must have a name.")
        if ("lifecycle:transition" not in xes_frame.columns
            or not xes_frame["lifecycle:transition"].isin(["complete", "start"]).all()
            or (len(xes_frame.loc[xes_frame["lifecycle:transition"] == "start"])
                != len(xes_frame.loc[xes_frame["lifecycle:transition"] == "complete"]))):
            raise XESIncorrectLifecycleTransitionError(
                "Each activity instance must have an event with a lifecycle transition"
                " from start to complete."
            )
        if ("time:timestamp" not in xes_frame.columns
            or xes_frame["time:timestamp"].isna().any()):
            raise XESMissingTimestamp("Each event must have a timestamp attribute.")
        if ("concept:name" not in xes_frame.columns
            or xes_frame["concept:name"].isna().any()):
            raise XESMissingActivityName("Each event must have an activity name attribute.")


def _validate_xes_dataframe_after_processing(xes_ground_truth: sf.FrameHE,
                                             xes_detected: sf.FrameHE) -> None:
    """ make sure that for each case-activity,
    the "lifecycle:transition" is always first 'start', then 'complete' """
    if (len(xes_detected["case:sampling_freq"].unique()) != 1
        or xes_detected["case:sampling_freq"].isna().any()):
        raise XESSamplingFreqError("The 'sampling_freq' attribute must have the same value for "
                                   "all traces in the detected log.")
    for xes_frame in [xes_ground_truth, xes_detected]:
        for case_id in xes_frame["case:concept:name"].unique():
            case = xes_frame.loc[xes_frame["case:concept:name"] == case_id]
            for activity_name in case["concept:name"].unique():
                activity = case.loc[case["concept:name"] == activity_name]
                current = "complete"
                for transition in activity["lifecycle:transition"].values:
                    if current == "complete" and transition == "start":
                        current = "start"
                    elif current == "start" and transition == "complete":
                        current = "complete"
                    else:
                        raise XESIncorrectLifecycleTransitionError(
                            "Each activity instance must have an event with a lifecycle transition"
                            " from start to complete."
                        )


def _validate_activity_name(event_log1_fr: sf.FrameHE,
                            event_log2_fr: sf.FrameHE,
                            activity_name: str) -> None:
    """Check if the activity name is in the event log.

    :param  event_log1_fr: The event log.
    :param  event_log2_fr: The event log.
    :param activity_name: The name of the activity to validate.
    :returns: None
    """
    if ((activity_name not in event_log1_fr["concept:name"].unique()
         and activity_name not in event_log2_fr["concept:name"].unique())
        and activity_name != "*"):
        raise ValueError(f"The activity name '{activity_name}' is not in the event logs.")


def _validate_case_id(event_log1_fr: sf.FrameHE,
                      event_log2_fr: sf.FrameHE,
                      case_id: str) -> None:
    """Check if the case ID is in the event log.

    :param event_log1_fr: The event log.
    :param event_log2_fr: The event log.
    :param str case_id: The case ID to validate.
    :returns: None
    """
    if ((case_id not in event_log1_fr["case:concept:name"].unique()
         or case_id not in event_log2_fr["case:concept:name"].unique())
        and case_id != "*"):
        raise ValueError(f"The case ID '{case_id}' is not in the event logs.")


def _logs_contain_at_most_one_case_id(event_log_one: sf.FrameHE,
                                      event_log_two: sf.FrameHE) -> bool:
    """Check if the event logs contains only one case ID.

    :param event_log_one: The event log.
    :param event_log_two: The event log.
    :returns: True if the event log contains only one case ID, False otherwise.
    :rtype: bool
    """
    all_case_names = (list(event_log_one["case:concept:name"].values)
                      + list(event_log_two["case:concept:name"].values))
    all_case_names = list(set(all_case_names))
    return len(all_case_names) == 0 or len(all_case_names) == 1


def _logs_contain_at_most_one_activity(event_log_one: sf.FrameHE,
                                       event_log_two: sf.FrameHE) -> bool:
    """Check if the event logs contain only one activity.

    :param event_log_one: The event log.
    :param event_log_two: The event log.
    :param str activity_name: The name of the activity to check for.
    :returns: True if the event logs contain only one activity, False otherwise.
    :rtype: bool
    """
    all_activity_names = (list(event_log_one["concept:name"].values)
                          + list(event_log_two["concept:name"].values))
    all_activity_names = list(set(all_activity_names))
    return len(all_activity_names) == 0 or len(all_activity_names) == 1


def _has_all_required_columns(event_log: sf.FrameHE) -> bool:
    """Check if the event log has all required columns.

    The required columns are ["case:concept:name", "concept:name",
    "lifecycle:transition", "time:timestamp"].
    :param event_log: The event log .
    :returns: True if the event log has all required columns, False otherwise.
    :rtype: bool
    """
    required_columns = ["case:concept:name", "concept:name",
                        "lifecycle:transition", "time:timestamp"]
    return all(col in list(event_log.columns) for col in required_columns)


def _case_activity_level_metric_pre_check(event_log_one: sf.FrameHE,
                                          event_log_two: sf.FrameHE) -> None:
    """Check if the event logs are in a valid format for case and activity level metrics.

    :param event_log_one: The event log.
    :param event_log_two: The event log.
    :param str activity_name: The name of the activity to validate.
    :param str case_id: The case ID to validate.
    :returns: None
    """
    if not _has_all_required_columns(event_log_one) or not _has_all_required_columns(event_log_two):
        raise ValueError(f"The event logs are missing required columns. {BUG_REPORT_CTA}")
    if not _logs_contain_at_most_one_case_id(event_log_one, event_log_two):
        raise ValueError(f"The event logs must contain exactly one case ID. {BUG_REPORT_CTA}")
    if not _logs_contain_at_most_one_activity(event_log_one, event_log_two):
        raise ValueError(f"The event logs must contain exactly one activity type. {BUG_REPORT_CTA}")


def _case_level_metric_pre_check(event_log_one: sf.FrameHE,
                                 event_log_two: sf.FrameHE) -> None:
    """Check if the event logs are in a valid format for case level metrics.

        :param event_log_one: The event log.
        :param event_log_two: The event log.
        :returns: None
        """
    if not _has_all_required_columns(event_log_one) or not _has_all_required_columns(event_log_two):
        raise ValueError(f"The event logs are missing required columns. {BUG_REPORT_CTA}")
    if not _logs_contain_at_most_one_case_id(event_log_one, event_log_two):
        raise ValueError(f"The event logs must contain exactly one case ID. {BUG_REPORT_CTA}")


def _determine_start_end_per_case(gt: sf.FrameHE,
                                  det: sf.FrameHE) -> sf.SeriesHE[sf.Index[np.str_], Any]:
    """Determine the start and end time per case.

    :param gt: The ground truth event log.
    :param det: The detected event log.
    :returns: A dictionary with the case ID as key and a tuple with the start and end time as value.
    """
    start_end_dict = {}
    for case_id in gt["case:concept:name"].unique():
        gt_case = gt.loc[gt["case:concept:name"] == case_id]
        det_case = det.loc[det["case:concept:name"] == case_id]
        if len(gt_case) == 0 or len(det_case) == 0:
            raise ValueError(f"Case ID '{case_id}' is not in both logs. "
                             "Currently, all case IDs must be in both logs. "
                             "Please clean logs before proceeding.")
        start_time = min(gt_case["time:timestamp"].min(), det_case["time:timestamp"].min())
        end_time = max(gt_case["time:timestamp"].max(), det_case["time:timestamp"].max())
        start_end_dict[case_id] = (start_time, end_time)
    return sf.SeriesHE.from_dict(start_end_dict)


# pylint: disable=too-many-positional-arguments, too-many-arguments, too-many-locals
def _generate_activity_metric_list(gt: sf.FrameHE,
                                   det: sf.FrameHE,
                                   case_id: str,
                                   activity_name: str,
                                   start_end_per_case: sf.SeriesHE[sf.Index[np.str_], Any],
                                   metric: Callable[[sf.FrameHE,
                                                     sf.FrameHE,
                                                     sf.SeriesHE,
                                                     float], Any],
                                   sampling_freq: float = -1.0) -> List[Any]:
    """ Generate a list of metrics for a given case and activity.

    :param gt: The ground truth log.
    :param det: The detected log.
    :param case_id: The case ID.
        If "*" is passed, the metric will be calculated and averaged for all
            case IDs.
    :param activity_name: The name of the activity.
        If "*" is passed, the metric will be calculated and averaged for all
            activities.
    :param start_end_per_case: The start and end times for each case.
    :param metric: The metric to calculate.
    :param sampling_freq: The sampling frequency of the logs, in Hertz.
    :returns: A list of metrics.
    """
    if not _has_all_required_columns(gt) or not _has_all_required_columns(det):
        raise ValueError("Logs must have columns case:concept:name, "
                         f"concept:name, lifecycle:transition, time:timestamp. {BUG_REPORT_CTA}")
    if case_id == "*":
        relevant_case_ids = list(set((list(gt["case:concept:name"].values)
                                      + list(det["case:concept:name"].values))))
    elif case_id != "*":
        relevant_case_ids = [case_id]
    else:
        raise ValueError(f"Case ID must be '*' or a valid case ID. {BUG_REPORT_CTA}")
    metric_list = []
    for case_name in relevant_case_ids:
        gt_filtered_by_case = gt.loc[gt["case:concept:name"] == case_name]
        det_filtered_by_case = det.loc[det["case:concept:name"] == case_name]
        if activity_name == "*":
            relevant_activity_names = list(
                set((list(gt_filtered_by_case["concept:name"].values)
                     + list(det_filtered_by_case["concept:name"].values))))
        elif activity_name != "*":
            relevant_activity_names = [activity_name]
        else:
            raise ValueError("Activity name must be '*' or a valid activity name. "
                             f"{BUG_REPORT_CTA}")
        for act_name in relevant_activity_names:
            gt_filtered_by_activity = gt_filtered_by_case.loc[gt_filtered_by_case[
                                                                  "concept:name"]
                                                              == act_name]
            det_filtered_by_activity = det_filtered_by_case.loc[det_filtered_by_case[
                                                                    "concept:name"]
                                                                == act_name]
            if len(gt_filtered_by_activity) == 0 and len(det_filtered_by_activity) == 0:
                continue
            metric_list.append(metric(gt_filtered_by_activity,
                                      det_filtered_by_activity,
                                      start_end_per_case,
                                      sampling_freq))
    if len(metric_list) == 0:
        raise ValueError(
            "No metrics could be calculated for this combination of case ID(s) and activity name("
            "s).")
    return metric_list


def _generate_sequence_metric_list(gt: sf.FrameHE,
                                   det: sf.FrameHE,
                                   case_id: str,
                                   metric: Callable[[sf.FrameHE,
                                                     sf.FrameHE], Any], ) -> List[Any]:
    """ Generate a list of sequence metrics for a given case.

    :param gt: The ground truth log.
    :param det: The detected log.
    :param case_id: The case ID to calculate the metric for.
        If "*" is passed, the metric will be calculated and averaged for all
            case IDs.
    :param metric: The metric to calculate.
    :returns: A list of metrics.
    """
    if not _has_all_required_columns(gt) or not _has_all_required_columns(det):
        raise ValueError("Logs must have columns case:concept:name, "
                         f"concept:name, lifecycle:transition, time:timestamp. {BUG_REPORT_CTA}")
    if case_id == "*":
        relevant_case_ids = list(set((list(gt["case:concept:name"].values)
                                      + list(det["case:concept:name"].values))))
    elif case_id != "*":
        relevant_case_ids = [case_id]
    else:
        raise ValueError(f"Case ID must be '*' or a valid case ID. {BUG_REPORT_CTA}")
    metric_list = []
    for case_name in relevant_case_ids:
        gt_filtered_by_case = gt.loc[gt["case:concept:name"] == case_name]
        det_filtered_by_case = det.loc[det["case:concept:name"] == case_name]
        if len(gt_filtered_by_case) == 0 and len(det_filtered_by_case) == 0:
            continue
        metric_list.append(metric(gt_filtered_by_case,
                                  det_filtered_by_case))
    return metric_list


def _count_values(input_frame: sf.FrameHE, column: str) -> Dict[Any, int]:
    """Count the number of unique values in a Series.

    :param input_frame: The input Series.
    :returns: A Series with the unique values as index and the count as values.
    """
    # Iterate over the rows of the input FrameHE and count the values
    # for the given column
    value_counts = {}
    for value in input_frame[column].values:
        if value not in value_counts:
            value_counts[value] = 1
        else:
            value_counts[value] += 1
    return value_counts


def _remove_events_with_length_zero(log: sf.FrameHE) -> sf.FrameHE:
    """Remove events with a duration of zero.

    :param log: The event log.
    :returns: The event log without events with a duration of zero.
    """
    case_activity_timestamp_combinations_to_remove = []
    for case_id in log["case:concept:name"].unique():
        case = log.loc[log["case:concept:name"] == case_id]
        for activity_name in case["concept:name"].unique():
            activity = case.loc[case["concept:name"] == activity_name]
            value_counts = _count_values(activity, "time:timestamp")
            keys = [key for key, value in value_counts.items() if value > 1]
            if len(keys) == 0:
                continue
            case_activity_timestamp_combinations_to_remove.extend(
                [(case_id, activity_name, key) for key in keys])
    if len(case_activity_timestamp_combinations_to_remove) == 0:
        return log
    for case_id, activity_name, timestamp in case_activity_timestamp_combinations_to_remove:
        log = log.loc[~((log["case:concept:name"] == case_id)
                        & (log["concept:name"] == activity_name)
                        & (log["time:timestamp"] == timestamp))]
    return log


def _get_case_in_filtered_logs(gt: sf.FrameHE, det: sf.FrameHE) -> str:
    """Get the case ID in the filtered logs.

    :param gt: The ground truth log.
    :param det: The detected log.
    :returns: None
    """

    if len(gt) > 0:
        case_id = str(gt["case:concept:name"].iloc[0])
    elif len(det) > 0:
        case_id = str(det["case:concept:name"].iloc[0])
    else:
        raise ValueError("Both logs, gt and det, are empty. "
                         f"Cannot calculate EA metrics. {BUG_REPORT_CTA}")
    return case_id



""" Tests for the cross correlation functionality of the aqudem package."""
import os
from unittest.mock import patch, MagicMock
import static_frame
import aqudem
from aqudem.cross_correlation_helper import _cross_correlation, _get_timeseries_format
from .mocks.logs import (basic_mock_gt, basic_mock_det,
                         start_end_series_basic,
                         start_end_series_basic_only_det,
                         basic_mock_gt_1, basic_mock_det_1,
                         start_end_series_basic_1,
                         start_end_series_basic_1_only_det,
                         start_end_series_basic_1_only_gt,
                         basic_mock_gt_2, basic_mock_det_2,
                         start_end_series_basic_2,)

EMPTY_FRAME_HE = static_frame.FrameHE.from_dict({
    "case:concept:name": [],
    "time:timestamp": [],
    "lifecycle:transition": [],
})

def test_get_timeseries_format_basic() -> None:
    res = _get_timeseries_format(basic_mock_gt,
                                 basic_mock_det,
                                 1,
                                 start_end_series_basic)
    assert res[0] == [1, 1, 1, 1, 1, 1, -1, -1, 1, 1, -1]
    assert res[1] == [-1, -1, -1, 1, 1, 1, 1, 1, -1, 1, 1]

    res = _get_timeseries_format(basic_mock_gt_1,
                                 basic_mock_det_1,
                                 1,
                                 start_end_series_basic_1)
    assert res[0] == [1, 1, 1, 1, 1, 1, -1, -1, -1, -1, -1]
    assert res[1] == [-1, -1, -1, -1, -1, 1, 1, 1, 1, 1, 1]

    res = _get_timeseries_format(basic_mock_gt_2,
                                 basic_mock_det_2,
                                 1,
                                 start_end_series_basic_2)
    assert res[0] == [1, 1, 1, 1, 1, 1, -1, -1, 1, 1, -1]
    assert res[1] == [-1, -1, -1, 1, 1, 1, 1, 1, -1, 1, 1]

def test_get_timeseries_format_basic_2hz() -> None:
    res = _get_timeseries_format(basic_mock_gt,
                                 basic_mock_det,
                                 2,
                                 start_end_series_basic)
    assert res[0] == [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                      -1, -1, -1, -1, -1, 1, 1, 1, -1, -1]
    assert res[1] == [-1, -1, -1, -1, -1, -1, 1, 1, 1, 1,
                      1, 1, 1, 1, 1, -1, -1, -1, 1, 1, 1]

def test_cc_basic() -> None:
    cc = _cross_correlation(basic_mock_gt,
                            basic_mock_det,
                            start_end_series_basic,
                            1)
    assert cc == (0.45, -0.2)


def test_cc_same_seq_returns_1_0() -> None:
    cc = _cross_correlation(basic_mock_det,
                            basic_mock_det,
                            start_end_series_basic_only_det,
                            1)
    assert cc == (1, 0.0)


def test_correct_sign_shift() -> None:
    cc = _cross_correlation(basic_mock_gt_1,
                            basic_mock_det_1,
                            start_end_series_basic_1,
                            1)
    assert cc == (0.55, -0.5)


def test_cc_none() -> None:
    """ Check that empty activity list leads to (None, None) return """
    cc1 = _cross_correlation(EMPTY_FRAME_HE,
                             basic_mock_det_1,
                             start_end_series_basic_1_only_det,
                             1)
    cc2 = _cross_correlation(basic_mock_gt_1,
                             EMPTY_FRAME_HE,
                             start_end_series_basic_1_only_gt,
                             1)
    cc3 = _cross_correlation(EMPTY_FRAME_HE,
                             EMPTY_FRAME_HE,
                             static_frame.SeriesHE.from_dict({}),
                             1)
    assert cc1 == (0.0, 0.0)
    assert cc2 == (0.0, 0.0)
    assert cc3 == (0.0, 0.0)


def test_cc_out_of_order() -> None:
    """ Basic check with out-of-order lists """
    cc = _cross_correlation(basic_mock_gt_2,
                            basic_mock_det_2,
                            start_end_series_basic_2,
                            1)
    assert cc == (0.45, -0.2)


@patch('aqudem.cross_correlation_helper._get_timeseries_format')
def test_cc_correct_behavior_one_activity(mock_get_timeseries_format: MagicMock) -> None:
    # Note that the function inputs are not used in the function,
    # because the mock is used to control the return value
    mock_get_timeseries_format.return_value = ([-1, -1, 1, -1, -1], [-1, -1, 1, -1, -1])
    cc = _cross_correlation(basic_mock_gt,
                            basic_mock_det,
                            start_end_series_basic,
                            1)
    assert cc == (1.0, 0.0)
    mock_get_timeseries_format.return_value = ([-1, -1, 1, 1, 1], [-1, -1, 1, 1, 1])
    cc = _cross_correlation(basic_mock_gt,
                            basic_mock_det,
                            start_end_series_basic,
                            1)
    assert cc == (1.0, 0.0)
    mock_get_timeseries_format.return_value = ([-1, -1, -1, -1, -1], [-1, -1, 1, -1, -1])
    cc = _cross_correlation(basic_mock_gt,
                            basic_mock_det,
                            start_end_series_basic,
                            1)
    assert cc == (0.6, 0.0)


@patch('aqudem.cross_correlation_helper._get_timeseries_format')
def test_cc_correct_shift_behavior(mock_get_timeseries_format: MagicMock) -> None:
    # Note that the function inputs are not used in the function,
    # because the mock is used to control the return value
    mock_get_timeseries_format.return_value = ([-1, -1, 1, -1, -1], [-1, -1, -1, 1, -1])
    cc = _cross_correlation(basic_mock_gt,
                            basic_mock_det,
                            start_end_series_basic,
                            1)
    assert cc == (0.8, -0.25)


def test_context_cross_correlation() -> None:
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.cross_correlation(activity_name="Activity A", case_id="ExampleTrace1")
    assert res == (0.73, 0.0)


def test_context_cross_correlation_by_activity() -> None:
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.cross_correlation(activity_name="Activity A")
    assert res == (0.64, 0.225)
    res = context.cross_correlation(activity_name="Activity B")
    assert res == (0.795, 0.165)
    res = context.cross_correlation(activity_name="Activity C")
    assert res == (0.83, -0.03)


def test_context_cross_correlation_by_case() -> None:
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.cross_correlation(case_id="ExampleTrace1")
    assert res == (0.8267, -0.01)
    res = context.cross_correlation(case_id="ExampleTrace2")
    assert res == (0.61, 0.39)


def test_context_cross_correlation_by_case_and_activity() -> None:
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    res = context.cross_correlation()
    assert res == (0.74, 0.15)



"""Functionality for the calculation of the Damerau-Levenshtein distance."""
from functools import partial
from typing import List, Union, Tuple
import static_frame as sf
import textdistance
from .utils import (_case_level_metric_pre_check, BUG_REPORT_CTA,
                    _generate_sequence_metric_list)


def _map_strings_to_letters(list1: List[str],
                            list2: Union[List[str], None] = None) -> Tuple[str, str]:
    """Create two strings where each letter represents a unique string from the list of strings.

    Example:
    (["i", "think", "i", "am", "think", "think", "am", "i"],
    ["therefore", "therefore", "am", "i", "i", "am", "therefore", "am"])
    -> ("ABACBBCA", "DDCAACDC")
    :param [str] list1: The first list of strings.
    :param [str] list2: The second list of strings.
    :return:
    """
    # Combine the two lists if the second list is provided
    all_strings = list1 + (list2 or [])

    # Create a mapping between unique strings and letters
    unique_strings = sorted(set(all_strings), key=all_strings.index)
    letter_mapping = {string: chr(ord('A') + i) for i, string in enumerate(unique_strings)}

    # Map strings to letters for the first list
    result1 = ''.join(letter_mapping[string] for string in list1)

    # Map strings to letters for the second list (if provided)
    result2 = ''
    if list2 is not None:
        result2 = ''.join(letter_mapping[string] for string in list2)

    return result1, result2


def _damerau_levenshtein_distancy_by_case(gt: sf.FrameHE,
                                          det: sf.FrameHE,
                                          case_id: str) -> Tuple[Union[int, float], float]:
    """Calculate the Damerau-Levenshtein distance between the ground truth and detected logs.

    :param gt: The ground truth log.
    :param det: The detected log.
    :param case_id: The case ID to calculate the Damerau-Levenshtein distance for.
        If "*" is passed, the Damerau-Levenshtein distance will be calculated and averaged for all
            case IDs.
    :return: The Damerau-Levenshtein distance and the normalized Damerau-Levenshtein distance.
    """
    all_d_l_distances = _generate_sequence_metric_list(gt, det, case_id,
                                                       partial(_damerau_opt_levenshtein_dist,
                                                               metr_type="dam_lev"))

    return (round(sum(dl[0] for dl in all_d_l_distances) / len(all_d_l_distances), 4),
            round(sum(dl[1] for dl in all_d_l_distances) / len(all_d_l_distances), 4))


def _levenshtein_distancy_by_case(gt: sf.FrameHE,
                                  det: sf.FrameHE,
                                  case_id: str) -> Tuple[Union[int, float], float]:
    """Calculate the Levenshtein distance between the ground truth and detected logs.

    :param gt: The ground truth log.
    :param det: The detected log.
    :param case_id: The case ID to calculate the Levenshtein distance for.
        If "*" is passed, the Levenshtein distance will be calculated and averaged for all
            case IDs.
    :return: The Levenshtein distance and the normalized Levenshtein distance.
    """
    all_d_l_distances = _generate_sequence_metric_list(gt, det, case_id,
                                                       partial(_damerau_opt_levenshtein_dist,
                                                               metr_type="lev"))

    return (round(sum(dl[0] for dl in all_d_l_distances) / len(all_d_l_distances), 4),
            round(sum(dl[1] for dl in all_d_l_distances) / len(all_d_l_distances), 4))


def _damerau_opt_levenshtein_dist(gt: sf.FrameHE,
                                  det: sf.FrameHE,
                                  metr_type: str = "dam_lev") -> Tuple[int, float]:
    """Calculate the (normalized) (Damerau-)Levenshtein distance
    between the ground truth and detected logs.

    Sorted based on start time.
    Assume the logs are filtered by case.
    :param gt: The ground truth log. Filtered by case_id.
    :param det: The detected log. Filtered by case_id.
    :return: Tuple of Damerau-Levenshtein distance and normed version.
    """
    _case_level_metric_pre_check(gt, det)
    gt_start_event_fr = (gt.loc[gt["lifecycle:transition"] == "start"]
                         .sort_values("lifecycle:transition"))
    dt_start_event_fr = (det.loc[det["lifecycle:transition"] == "start"]
                         .sort_values("lifecycle:transition"))
    gt_activity_names = list(gt_start_event_fr["concept:name"].values)
    dt_activity_names = list(dt_start_event_fr["concept:name"].values)
    gt_string, dt_string = _map_strings_to_letters(gt_activity_names, dt_activity_names)
    max_len = max(len(gt_string), len(dt_string))
    if max_len == 0:
        return (0, 0.0)
    if metr_type == "lev":
        return (textdistance.levenshtein(gt_string, dt_string),
                textdistance.levenshtein(gt_string, dt_string) / max_len)
    if metr_type == "dam_lev":
        return (textdistance.damerau_levenshtein(gt_string, dt_string),
                textdistance.damerau_levenshtein(gt_string, dt_string) / max_len)
    raise ValueError(f"Type must be 'dam_lev' or 'lev'. {BUG_REPORT_CTA}")



""" This module contains the cross correlation helper functions. """
import math
from datetime import datetime, timedelta
from typing import List, Tuple, Any
import static_frame as sf
import numpy as np
from .utils import _generate_activity_metric_list


# pylint: disable=too-many-positional-arguments, too-many-arguments
def _cross_correlation_by_activity_case(ground_truth: sf.FrameHE,
                                        detected: sf.FrameHE,
                                        sampling_freq: float,
                                        activity_name: str,
                                        case_id: str,
                                        start_end_per_case: \
                                            sf.SeriesHE[sf.Index[np.str_], Any]) \
                                                -> Tuple[float, float]:
    """Calculate the cross correlation between the ground truth and detected logs.

    :param ground_truth: The ground truth log.
    :param detected: The detected log.
    :param sampling_freq: The sampling frequency of the logs, in Hertz.
    :param activity_name: The name of the activity.
        If "*" is passed, the cross-correlation will be calculated and averaged for all activities.
    :param case_id: The case ID.
           If "*" is passed, the cross-correlation will be calculated and averaged for all case IDs.
    :param start_end_per_case: The start and end times for each case.
    :return: The cross correlation.
    """
    cross_correlations = _generate_activity_metric_list(gt=ground_truth,
                                                        det=detected,
                                                        case_id=case_id,
                                                        activity_name=activity_name,
                                                        start_end_per_case=start_end_per_case,
                                                        metric=_cross_correlation,
                                                        sampling_freq=sampling_freq)
    return (round(sum(cc[0] for cc in cross_correlations) / len(cross_correlations), 4),
            round(sum(cc[1] for cc in cross_correlations) / len(cross_correlations), 4))


def _cross_correlation(gt: sf.FrameHE,
                       det: sf.FrameHE,
                       start_end_per_case: sf.SeriesHE[sf.Index[np.str_], Any],
                       sampling_rate_hz: float) -> Tuple[float, float]:
    """
    Relative shift > 0 -> the detected activity timeseries had
    to be pushed forward in time (time delay/detected later than actual)
    Relative shift < 0 -> the detected activity timeseries had
    to be pulled back (time advance/detected earlier than actual)

    Assume
    :param gt: Ground truth log
    :param det: Detected log
    :param start_end_per_case: Start and end times for all cases
    :param sampling_rate_hz: Sampling rate in Hz
    :return: cross correlation and relative shift necessary for max cc
    """
    if len(gt) == 0 or len(det) == 0:
        return 0.0, 0.0
    time_series_bin_gt, time_series_bin_det = _get_timeseries_format(gt,
                                                                     det,
                                                                     sampling_rate_hz,
                                                                     start_end_per_case)
    a = np.array(time_series_bin_gt)
    b = np.array(time_series_bin_det)
    if len(a) != len(b):
        raise ValueError("Time series must have same length.")
    normalization_factor = len(a)
    c = np.correlate(a, b, mode='full')
    c_max = c.max()
    c_list = list(c)
    if c_max != 0.0:
        c_ind = c_list.index(c_max)
        shift_total = c_ind - math.floor(len(c_list) / 2)
        shift_relative = shift_total / math.floor(len(c_list) / 2)
    else:
        shift_relative = 0.0
    return round(c_max / normalization_factor, 2), round(shift_relative, 2)


def _get_timeseries_format(gt: sf.FrameHE,
                           det: sf.FrameHE,
                           sampling_rate_hz: float,
                           start_end_per_case: sf.SeriesHE[sf.Index[np.str_], Any], ) \
    -> Tuple[List[int], List[int]]:
    """
    Turns a list of ground truth and detected activity
    instances into a time series when the activity is running and
    when not. All activities must be of same type (have same name).
    :param gt:
    :param det:
    :param sampling_rate_hz:
    :return:
    """
    case_id = gt["case:concept:name"].iloc[0]
    sorted_gt = gt.sort_values("time:timestamp")
    sorted_det = det.sort_values("time:timestamp")
    min_start: datetime = start_end_per_case[case_id][0]
    max_stop: datetime = start_end_per_case[case_id][1]
    time_step = timedelta(seconds=1) / sampling_rate_hz
    timeseries_gt = []
    timeseries_det = []

    current_time = min_start
    while current_time <= max_stop:
        rel_rows_gt = sorted_gt.loc[(sorted_gt["time:timestamp"] >= current_time)]
        if len(rel_rows_gt) > 0 and (rel_rows_gt["lifecycle:transition"].
                                         values[0] == "complete"
                                     or (rel_rows_gt["time:timestamp"]
                                             .values[0] == current_time and
                                         rel_rows_gt["lifecycle:transition"]
                                             .values[0] == "start")):
            timeseries_gt.append(1)
        else:
            timeseries_gt.append(-1)
        rel_rows_det = sorted_det.loc[(sorted_det["time:timestamp"] >= current_time)]
        if len(rel_rows_det) > 0 and (rel_rows_det["lifecycle:transition"]
                                          .values[0] == "complete"
                                        or (rel_rows_det["time:timestamp"]
                                                .values[0] == current_time and
                                            rel_rows_det["lifecycle:transition"]
                                                .values[0] == "start")):
            timeseries_det.append(1)
        else:
            timeseries_det.append(-1)
        current_time += time_step
    return timeseries_gt, timeseries_det



"""Helper functions for metrics from Ward et al; used in event analysis and 2SET."""
from datetime import datetime
from functools import lru_cache
from typing import Union
import static_frame as sf
from numpy import nan
from .utils import BUG_REPORT_CTA

EIGHT_TYPE_MAPPING_P_C_N = {
    (nan, "FP", "TN"): "I",
    (nan, "FP", "FN"): "I",
    (nan, "FN", "TN"): "D",
    (nan, "FN", "FP"): "D",
    (nan, "FP", "TP"): "Oa",
    (nan, "FN", "TP"): "Ua",
    ("TP", "FP", "TP"): "M",
    ("TP", "FN", "TP"): "F",
    ("TN", "FP", "TP"): "Oa",
    ("FN", "FP", "TP"): "Oa",
    ("TN", "FN", "TP"): "Ua",
    ("FP", "FN", "TP"): "Ua",
    ("TN", "FP", "TN"): "I",
    ("FN", "FP", "TN"): "I",
    ("TN", "FP", "FN"): "I",
    ("FN", "FP", "FN"): "I",
    ("TN", "FN", "TN"): "D",
    ("FP", "FN", "TN"): "D",
    ("TN", "FN", "FP"): "D",
    ("FP", "FN", "FP"): "D",
    ("TP", "FP", "TN"): "Oo",
    ("TP", "FP", "FN"): "Oo",
    ("TP", "FN", "TN"): "Uo",
    ("TP", "FN", "FP"): "Uo",
    ("TN", "FP", nan): "I",
    ("FN", "FP", nan): "I",
    ("TN", "FN", nan): "D",
    ("FP", "FN", nan): "D",
    ("TP", "FP", nan): "Oo",
    ("TP", "FN", nan): "Uo"
}

FOUR_TYPE_MAPPING_GT_DET = {
    (True, True): "TP",
    (False, False): "TN",
    (False, True): "FP",
    (True, False): "FN"
}


def _generate_eight_type(prev_type: Union[str, None],
                         segment_type: str,
                         next_type: Union[str, None]) -> str:
    """Generate the eight type of the segment based on the
    type of the segment and the previous and next segment.

    :param str segment_type: The type of the segment.
        In ["TP", "TN", "I", "M", "O", "D", "F", "U"].
    :param str prev_type: The type of the previous segment.
        In ["TP", "TN", "I", "M", "O", "D", "F", "U"].
    :param str next_type: The type of the next segment.
        In ["TP", "TN", "I", "M", "O", "D", "F", "U"].
    :return: The eight type of the segment.
    :rtype: str
    """
    if segment_type not in ["TP", "TN", "FP", "FN"]:
        raise ValueError("The segment type must be 'TP', 'TN', 'FP', or 'FN'.")
    if prev_type not in ["TP", "TN", "FP", "FN", nan]:
        raise ValueError("The previous segment type must be 'TP', 'TN', 'FP', 'FN', or None.")
    if next_type not in ["TP", "TN", "FP", "FN", nan]:
        raise ValueError("The next segment type must be 'TP', 'TN', 'FP', 'FN', or None.")
    if segment_type in ["TP", "TN"]:
        return segment_type
    return EIGHT_TYPE_MAPPING_P_C_N[(prev_type, segment_type, next_type)]


def _is_during_activity_exec(log: sf.FrameHE, timestamp: datetime) -> bool:
    """ Check if the timestamp is during an activity execution.

    Assumes event log to be filtered for certain activity and case.
    Relevant columns/format of the input DataFrames:
    ground_truth:
        - "lifecycle:transition": The lifecycle transition of the event.
        - "time:timestamp": The timestamp of the event, start and complete for all
    :param pd.DataFrame log: The event log.
    :param datetime timestamp: The timestamp to check.
    """
    # check if middle timestamp is during activity in ground truth and detected
    # get the first value with a timestamp smaller than the middle timestamp
    log_before_middle = log.loc[log["time:timestamp"] <= timestamp]
    if len(log_before_middle) == 0:
        return False
    highest_before_middle = log_before_middle.iloc[-1]
    if highest_before_middle["lifecycle:transition"] == "start":
        return True
    if highest_before_middle["lifecycle:transition"] == "complete":
        return False
    raise ValueError("The ground truth log is not in the correct format.")


# pylint: disable=too-many-locals
@lru_cache
def _generate_segment_scores(ground_truth: sf.FrameHE,
                             detected: sf.FrameHE,
                             start_time: datetime,
                             end_time: datetime) -> sf.FrameHE:
    """Generate the segment scores for the ground truth and detected logs.

    Assumes event logs to be filtered for certain activity and case.
    Relevant columns/format of the input DataFrames:
    ground_truth:
        - "lifecycle:transition": The lifecycle transition of the event.
        - "time:timestamp": The timestamp of the event.
    detected:
        - "lifecycle:transition": The lifecycle transition of the event.
        - "time:timestamp": The timestamp of the event.
    Format of the generated DataFrame:
    Columns:
        - "segment_id": The segment ID.
        - "start": The start timestamp of the segment.
        - "end": The end timestamp of the segment.
        - "type": The type of the segment.
            In ["TP", "TN", "I", "M", "Oa", "Oo", "D", "F", "Ua", "Uo"].
    :param ground_truth: The ground truth log.
    :param detected: The detected log.
    :param start_time: The start time of the trace/case.
    :param end_time: The end time of the trace/case.
    """
    if len(ground_truth) == 0 and len(detected) == 0:
        return sf.FrameHE(columns=["start", "end", "type"])
    # Filter the logs to only contain the start and complete events
    ground_truth_filtered: sf.FrameHE = ground_truth.loc[
        ground_truth["lifecycle:transition"].isin(["start", "complete"])]
    detected_filtered = detected.loc[detected["lifecycle:transition"].isin(["start", "complete"])]
    # merge the logs and extract list so that
    # every start and complete event are turned into a boundary
    merged = sf.FrameHE.from_concat([ground_truth_filtered, detected_filtered],
                                    index=sf.IndexAutoFactory,
                                    axis=0)
    merged = merged.sort_values("time:timestamp")
    boundary_timestamps = list(set(list(merged["time:timestamp"].values)))
    boundary_timestamps.sort()
    if start_time < boundary_timestamps[0]:
        boundary_timestamps.insert(0, start_time)
    elif start_time > boundary_timestamps[0]:
        raise ValueError(f"Internal value error regarding timing. {BUG_REPORT_CTA}")
    if end_time > boundary_timestamps[-1]:
        boundary_timestamps.append(end_time)
    elif end_time < boundary_timestamps[-1]:
        raise ValueError(f"Internal value error regarding timing. {BUG_REPORT_CTA}")
    # create a new Dataframe with columns start, end, type
    segment_scores_list = []
    # make sure that the boundary timestamps are in the correct order
    if boundary_timestamps != sorted(boundary_timestamps):
        raise ValueError(f"The boundary timestamps are not in the correct order. {BUG_REPORT_CTA}")
    # iterate over the boundary timestamps and create the segments
    current_type = "None"
    for segment in range(len(boundary_timestamps) - 1):
        start = boundary_timestamps[segment]
        end = boundary_timestamps[segment + 1]
        # get the middle timestamp of the segment
        middle = start + (end - start) / 2
        gt_is_active = _is_during_activity_exec(ground_truth_filtered, middle)
        det_is_active = _is_during_activity_exec(detected_filtered, middle)
        segment_type = FOUR_TYPE_MAPPING_GT_DET[(gt_is_active, det_is_active)]
        # append the segment to the segment_scores DataFrame
        prior_type = current_type
        current_type = segment_type
        if prior_type == current_type:
            raise ValueError(f"The current and prior segment types are the same. {BUG_REPORT_CTA}")
        segment_scores_list.append({"start": start, "end": end, "type": segment_type})
    segment_scores_frame = sf.FrameHE.from_dict_records(segment_scores_list)
    # sort the segment_scores DataFrame by the start timestamp
    segment_scores_frame = segment_scores_frame.sort_values("start")
    # extend the dataframe so that for every row there are new columns
    # prev_type and next_type based on the type of the previous and next segment
    segment_scores_frame = (
        sf.FrameHE.from_concat([segment_scores_frame, segment_scores_frame["type"].shift(1)],
                               axis=1,
                               columns=['start', 'end', 'type', 'prev_type']))
    segment_scores_frame = (
        sf.FrameHE.from_concat([segment_scores_frame, segment_scores_frame["type"].shift(-1)],
                               axis=1,
                               columns=['start', 'end', 'type', 'prev_type', 'next_type']))
    # generate new column eight_type based on the type
    # of the segment and the previous and next segment
    eight_types = (
        segment_scores_frame.iter_series(axis=1).apply(
            lambda x: _generate_eight_type(x["prev_type"], x["type"], x["next_type"])))
    segment_scores_frame = (
        sf.FrameHE.from_concat([segment_scores_frame, eight_types],
                               axis=1,
                               columns=['start', 'end', 'type', 'prev_type',
                                        'next_type', 'eight_type']))
    # remove type, prev_type, and next_type columns
    segment_scores_frame = segment_scores_frame.drop[["type", "prev_type", "next_type"]]
    # rename the eight_type column to type
    segment_scores_frame = segment_scores_frame.relabel(columns=['start', 'end', 'type'])
    return segment_scores_frame



"""Tests for `aqudem` package."""
import os
import aqudem


def test_context_creation() -> None:
    """Test the creation of a context object."""
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    for value in context._ground_truth["concept:name"].values:
        assert value in ["Activity A", "Activity B", "Activity C"]
    for value in context._detected["concept:name"].values:
        assert value in ["Activity A", "Activity B", "Activity C"]
    for value in context._detected["case:sampling_freq"].values:
        assert value == 1.0
    assert context._ground_truth["case:concept:name"].count(unique=True) == 2
    assert context._detected["case:concept:name"].count(unique=True) == 2
    assert len(context._ground_truth.loc[
                   context._ground_truth["lifecycle:transition"] == "complete"]) == 9
    assert len(context._ground_truth.loc[
                   context._ground_truth["lifecycle:transition"] == "start"]) == 9
    assert len(context._detected.loc[
                   context._detected["lifecycle:transition"] == "complete"]) == 8
    assert len(context._detected.loc[
                   context._detected["lifecycle:transition"] == "start"]) == 8


def test_context_creation_multiple_traces() -> None:
    """Test the creation of a context object."""
    aqudem.Context(os.path.join("tests", "resources", "ground_truth_multipletraces.xes"),
                   os.path.join("tests", "resources", "detected_multipletraces.xes"))


def test_get_activity_names() -> None:
    """Test the get_activity_names method."""
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth.xes"),
                             os.path.join("tests", "resources", "detected.xes"))
    act_names = context.activity_names
    assert "Activity A" in act_names["ground_truth"]
    assert "Activity B" in act_names["ground_truth"]
    assert "Activity C" in act_names["ground_truth"]
    assert "Activity A" in act_names["detected"]
    assert "Activity B" in act_names["detected"]
    assert "Activity C" in act_names["detected"]
    assert len(act_names["ground_truth"]) == 3
    assert len(act_names["detected"]) == 3


def test_get_case_ids() -> None:
    """Test the get_case_ids method."""
    context = aqudem.Context(os.path.join("tests", "resources", "ground_truth_multipletraces.xes"),
                             os.path.join("tests", "resources", "detected_multipletraces.xes"))
    case_ids = context.case_ids
    assert isinstance(case_ids, dict)
    assert ["ground_truth", "detected"] == list(case_ids.keys())
    assert "ExampleTrace" in case_ids["ground_truth"]
    assert "ExampleTrace2" in case_ids["ground_truth"]
    assert "ExampleTrace" in case_ids["detected"]
    assert "ExampleTrace2" in case_ids["detected"]
    assert len(case_ids["ground_truth"]) == 2
    assert len(case_ids["detected"]) == 2
